/*
		Vertex Shader
		This shader will convert the incoming VBO to texture
		This is the skinned-mesh version using matrices and bones
		The skinning is done similiar to the NVIDIA OpenGL SDK Skinning example.

		Matthias Hollnder 2008/2009
*/

// --------------------------------------------------
// shader definition
// --------------------------------------------------
#version 120

// --------------------------------------------------
// Attributes
// --------------------------------------------------
attribute vec4 			vWeight;				// weights for the bones
attribute vec4 			vIndex;					// indices of bones
attribute float 		fNumBones;				// number of bones that are connected to this vertex
attribute float			fVertexID;				// this is like gl_VertexID which doesn't exist without VBO

// --------------------------------------------------
// uniform variables:
// --------------------------------------------------
uniform int				iOutputSamplerSize;
uniform mat4 			MatBones[30];

// --------------------------------------------------
// varying variables
// --------------------------------------------------
varying vec3			vVarColorVertex;					// The color/value (=newVertex)
varying vec3			vVarColorNormal;

// --------------------------------------------------
// function to convert output position
// --------------------------------------------------
vec2 ConvertOutputPosition(in unsigned int uiIndex, in int iWrapSize)
{
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiIndex / uiWrap;
		
		float fX = float(uiIndex - (uiY*uiWrap));
		float fY = float(uiY);
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// Vertex-Shader
// --------------------------------------------------
void main() {
		// This will be the resulting position:
		vec4 vFinalVertex = vec4(0.0, 0.0, 0.0, 0.0);
		vec3 vFinalNormal = vec3(0.0, 0.0, 0.0);

		// Those two are used for shuffling
		vec4 vCurIndex = vIndex;
		vec4 vCurWeight = vWeight;

		// run through all assigned bones:
		for (int i=0; i< int(fNumBones); ++i)
		{
				mat4 m44 = MatBones[int(vCurIndex.x)];
				
				// transform the offset by bone i
				vFinalVertex += m44 * gl_Vertex * vCurWeight.x;

				mat3 m33 = mat3(m44[0].xyz,
								m44[1].xyz,
								m44[2].xyz);

				// transform normal by bone i
				vFinalNormal += m33 * gl_Normal * vCurWeight.x;

				// shift over the index/weight variables, this moves the index and 
				// weight for the current bone into the .x component of the index 
				// and weight variables
				vCurIndex = vCurIndex.yzwx;
				vCurWeight = vCurWeight.yzwx;
		}

		// calculate vertex position
		unsigned int uiVertexID = unsigned int( fVertexID );					// Vertex (ID) to process
		vec2 vPos2D = ConvertOutputPosition( uiVertexID, iOutputSamplerSize );	// The position to write to
		vec4 vPos = vec4( vPos2D.xy, 0.0, 1.0 );								// z=0.0, w=1.0

		gl_Position = ( gl_ModelViewProjectionMatrix * vPos );

		// The color is the position
		vVarColorVertex = vFinalVertex.xyz;
		vVarColorNormal = vFinalNormal.xyz;
}
