/*
	Matthias Hollnder & Tamy Boubekeur
	Copyright (c) 2010 Telecom ParisTech
*/

// --------------------------------------------------
// shader definition
// --------------------------------------------------
#version 120

// --------------------------------------------------
// Uniform variables:
// --------------------------------------------------
uniform int 			iLightingModel;						// A value to control the specular lighting model
uniform float 			fSurfaceSmoothness;

// --------------------------------------------------
// varying variables:
// --------------------------------------------------
varying vec3 			V;									// Vertex in View-Space
varying vec3 			N;									// Normal of the Vertex
varying vec3 			vLightDir;							// VecToLight

// --------------------------------------------------
// consts used for lighting
// --------------------------------------------------
const float fe = 2.71828182845904523536;					// Euler number

// --------------------------------------------------
// Helper function to retrieve angle between
// two NORMALIZED vectors in Rad
// --------------------------------------------------
float GetDiffAngle(in vec3 v1, in vec3 v2)
{
		return acos(dot(v1, v2));
}

// --------------------------------------------------
// Ambient term
// --------------------------------------------------
vec4 CalcAmbient(in gl_LightModelParameters vLightDef,
				 in gl_MaterialParameters vMatDef)
{
		return vLightDef.ambient * vMatDef.ambient;
}

// --------------------------------------------------
// Diffuse term
// --------------------------------------------------
vec4 CalcDiffuse(in gl_LightSourceParameters vLightDef,
				 in gl_MaterialParameters vMatDef,
				 in vec3 vLightVec,
				 in vec3 vNormal)
{
		float fLambert = max(dot(vLightVec, vNormal), 0.0);
		return fLambert * (vLightDef.diffuse * vMatDef.diffuse);
}

// --------------------------------------------------
// Specular term (Phong)
// --------------------------------------------------
vec4 CalcSpecularPhong(in gl_LightSourceParameters vLightDef,
				       in gl_MaterialParameters vMatDef,
					   in vec3 vLightVec,
					   in vec3 vNormal,
					   in vec3 vVertex)
{
		vec3 vRefl = reflect(vLightVec, vNormal);				// reflected Vector
		float fSpec = pow(max(dot(vRefl, vVertex), 0.0), vMatDef.shininess);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Specular term (Blinn)
// --------------------------------------------------
vec4 CalcSpecularBlinn(in gl_LightSourceParameters vLightDef,
				       in gl_MaterialParameters vMatDef,
					   in vec3 vLightVec,
					   in vec3 vNormal,
					   in vec3 vVertex)
{
		vec3 vHalf = normalize((vLightVec + -vVertex)*0.5);
		float fSpec = pow(max(dot(vNormal, vHalf), 0.0), vMatDef.shininess);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Specular term (Gaussian)
// --------------------------------------------------
vec4 CalcSpecularGaussian(in gl_LightSourceParameters vLightDef,
						  in gl_MaterialParameters vMatDef,
						  in vec3 vLightVec,
						  in vec3 vNormal,
						  in vec3 vVertex)
{
		vec3 vHalf = normalize((vLightVec + -vVertex)*0.5);
		float fFraction = (GetDiffAngle(vNormal,vHalf)) / fSurfaceSmoothness;
		float fSpec = pow(fe, - (fFraction*fFraction));
		fSpec = max(fSpec, 0.0);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Specular term (Beckmann)
// --------------------------------------------------
vec4 CalcSpecularBeckmann(in gl_LightSourceParameters vLightDef,
						  in gl_MaterialParameters vMatDef,
						  in vec3 vLightVec,
						  in vec3 vNormal,
						  in vec3 vVertex)
{
		vec3 vHalf = normalize((vLightVec + -vVertex)*0.5);
		float fCosTerm = dot(vNormal, vHalf);
		float fDivisor = 4.0 * fSurfaceSmoothness * fSurfaceSmoothness * pow(fCosTerm,4.0);
		float fTan = tan(GetDiffAngle(vNormal, vHalf));
		float fExponent = fTan / fSurfaceSmoothness;
		fExponent *= fExponent;
		float fSpec = (pow(fe, -fExponent)) / fDivisor;
		fSpec = max(fSpec, 0.0);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Specular term (Cook-Torrance)
// --------------------------------------------------
vec4 CalcSpecularCookTorrance(in gl_LightSourceParameters vLightDef,
						      in gl_MaterialParameters vMatDef,
							  in vec3 vLightVec,
							  in vec3 vNormal,
							  in vec3 vVertex)
{
		// Lambda for the Fresnel-Term
		float fLambda = 0.8;	//vMatDef.shininess;
	
		vec3 E = -vVertex;
		float fEDotN = dot(E, vNormal);
		vec3 vHalf = normalize( vLightVec - vVertex );
		float fNDotH = dot(vNormal, vHalf);		

		// calculate D:
		float fDivisor = 4.0*fSurfaceSmoothness*fSurfaceSmoothness*fNDotH*fNDotH*fNDotH*fNDotH;
		float fExponent = tan(GetDiffAngle(vNormal, vHalf)) / fSurfaceSmoothness;
		fExponent *= fExponent;
		float fD = (pow(fe, -fExponent)) / fDivisor;
		
		// calculate F (using the Schlick-term):
		//float fF = pow((1.0 + fEDotN), fLambda);		// that looks so bad!!!
		float fF = fLambda + ((1.0-fLambda)*pow(1.0 - dot(E, vHalf),5.0));
		
		// calculate G:
		float fEDotH_Rec = 1.0 / dot(E, vHalf);
		float fG1 = 2.0*fNDotH*fEDotN*fEDotH_Rec;
		float fG2 = 2.0*fNDotH*(max(dot(vLightVec, vNormal),0.0))*fEDotH_Rec;
		float fG = min(max(min(fG1, fG2), 0.0), 1.0);
		
		// all in all:
		float fSpec = (fD*fF*fG)/(fEDotN);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Specular term (Christophe Schlick)
// --------------------------------------------------
vec4 CalcSpecularSchlick(in gl_LightSourceParameters vLightDef,
						 in gl_MaterialParameters vMatDef,
						 in vec3 vLightVec,
						 in vec3 vNormal,
						 in vec3 vVertex)
{
		vec3 vRefl = reflect(vLightVec, vNormal);
		float fT = max(dot(vRefl, vVertex), 0.0);
		float fSpec = max(fT / (vMatDef.shininess - vMatDef.shininess*fT + fT), 0.0);
		return fSpec * (vLightDef.specular * vMatDef.specular);
}

// --------------------------------------------------
// Attenuation factor
// --------------------------------------------------
float CalcAttenuation(in gl_LightSourceParameters vLightDef,
				      in vec3 vVertex,
					  in vec3 vLightVec)
{
		float fDistance = length(vVertex - vLightVec);
		return (1.0 / (vLightDef.constantAttenuation +
					   vLightDef.linearAttenuation * fDistance +
					   vLightDef.quadraticAttenuation * (fDistance * fDistance)));
}

// --------------------------------------------------
// Fragment Shader:
// --------------------------------------------------
void main() {
		vec4 vSpecular  = vec4(0.0, 0.0, 0.0, 0.0);

		vec3 N_Normed = normalize(N);									// Normal
		
		// Calculate Attenuation before the LightVec is normalized
		float fAttenuation = CalcAttenuation(gl_LightSource[0],
											 V,
											 vLightDir);

		vec3 vLightDir_Normed = normalize(vLightDir);
		vec3 V_Normed = normalize(V);
		
		vec4 vDiffuse = CalcDiffuse(gl_LightSource[0],
								    gl_FrontMaterial,
									vLightDir_Normed,
									N_Normed);
		vec4 vAmbient = CalcAmbient(gl_LightModel,
								    gl_FrontMaterial);
									
		if (gl_FrontMaterial.shininess <= 0.00001) {
				vSpecular = vec4(0.0, 0.0, 0.0, 1.0);
		}
		else if (iLightingModel == 0) {
				vSpecular = CalcSpecularPhong(gl_LightSource[0],
										      gl_FrontMaterial,
											  vLightDir_Normed,
											  N_Normed,
											  V_Normed);
		}
		else if (iLightingModel == 1) {
				vSpecular = CalcSpecularBlinn(gl_LightSource[0],
										      gl_FrontMaterial,
											  vLightDir_Normed,
											  N_Normed,
											  V_Normed);
		}
		else if (iLightingModel == 2) {
				vSpecular = CalcSpecularGaussian(gl_LightSource[0],
												 gl_FrontMaterial,
												 vLightDir_Normed,
												 N_Normed,
												 V_Normed);
		}
		else if (iLightingModel == 3) {
				vSpecular = CalcSpecularBeckmann(gl_LightSource[0],
												 gl_FrontMaterial,
												 vLightDir_Normed,
												 N_Normed,
												 V_Normed);
		}
		else if (iLightingModel == 4) {
				vSpecular = CalcSpecularCookTorrance(gl_LightSource[0],
												     gl_FrontMaterial,
													 vLightDir_Normed,
													 N_Normed,
													 V_Normed);
		}
		else {
				vSpecular = CalcSpecularSchlick(gl_LightSource[0],
												gl_FrontMaterial,
												vLightDir_Normed,
												N_Normed,
												V_Normed);
		}
		
	
		gl_FragColor = (fAttenuation * (vAmbient + vDiffuse + vSpecular));
}
