/*
		This shader will perform the first subdivision step
		on the GPU using a weighting texture.
		This shader is used for computing the normals of the mesh
		
		Matthias Hollnder 2008/2009
*/

// --------------------------------------------------
// shader definition
// --------------------------------------------------
#version 120
#extension GL_EXT_gpu_shader4 : enable

// --------------------------------------------------
// uniform variables:
// --------------------------------------------------
uniform int				CalculationSamplerSize;
uniform sampler2D		CalculationSampler;					// this is the texture used for performing the 1st Subdivision on the GPU
uniform int				iNormalStride;						// how many vec4 are for each vertex

uniform int				iInputSamplerSize;					// The size of the InputSamplers
uniform sampler2D		InputSamplerVertices;				// vertices

uniform int				iOutputSamplerSize;

// --------------------------------------------------
// varying variables
// --------------------------------------------------
varying vec3			vVarColor;

// --------------------------------------------------
// const
// --------------------------------------------------
const int 				ciMaxValence = 20;					// Max valence allowed for this shader (must be devidable by 4)

// --------------------------------------------------
// function to convert 1D-index to 2D-index
// --------------------------------------------------
vec2 Convert1DIndexTo2DIndex(in unsigned int uiVertexIndex, in int iWrapSize)
{
		float fWrapRec = 1.0 / float(iWrapSize);
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiVertexIndex / uiWrap;
		unsigned int uiX = uiVertexIndex - (uiY* uiWrap);
		
		float fX = float(uiX) * fWrapRec;
		float fY = float(uiY) * fWrapRec;
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// function to convert output position
// --------------------------------------------------
vec2 ConvertOutputPosition(in unsigned int uiIndex, in int iWrapSize)
{
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiIndex / uiWrap;
		
		float fX = float(uiIndex - (uiY*uiWrap));
		float fY = float(uiY);
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// Loading a vertex from texture by index
// --------------------------------------------------
vec3 LoadVertex( unsigned int uiVertexID )
{
		vec2 vTexCoord = Convert1DIndexTo2DIndex( uiVertexID, iInputSamplerSize );
		return texture2D( InputSamplerVertices, vTexCoord ).xyz;
}

// --------------------------------------------------
// calculating the normal
// --------------------------------------------------
vec3[ciMaxValence] LoadOneRingOfVertex( in unsigned int uiVertexID, out int iValence, out bool bIsInterior )
{
		vec3 vReturnArray[ciMaxValence];
		
		for (int i=0; i< ciMaxValence; ++i) {
				vReturnArray[i] = vec3( 0, 0, 0 );
		}
		
		
		vec2 vTexCoord = vec2( 0, 0 );
		vec4 vCalcVals = vec4( 0, 0, 0, 0 );
		
		unsigned int uiIdx = uiVertexID * unsigned int( iNormalStride );										// Index in CalculationTexture
		vTexCoord = Convert1DIndexTo2DIndex( uiIdx, CalculationSamplerSize );
		vCalcVals = texture2D( CalculationSampler, vTexCoord ).xyzw;
		
		bIsInterior = vCalcVals.x > -1.3;																		// All interiot vertices have a value of -1
		iValence = int( vCalcVals.y );																			// one-ring-size in y
		
		vReturnArray[0] = LoadVertex( unsigned int( vCalcVals.z ) );											// Loading two vertices since 4 components where used
		vReturnArray[1] = LoadVertex( unsigned int( vCalcVals.w ) );
		int iCurrentVertex = 2;
		
		bool bHasFinished = false;
		
		for (int i=1; i < ciMaxValence / 4; ++i) {
				vTexCoord = Convert1DIndexTo2DIndex( uiIdx + unsigned int( i ), CalculationSamplerSize );
				vCalcVals = texture2D( CalculationSampler, vTexCoord ).xyzw;
				
				for (int j=0; j<4; ++j) {
						if ( vCalcVals.x > -0.5 ) {
								vReturnArray[ iCurrentVertex ] = LoadVertex( unsigned int (vCalcVals.x) );
								++iCurrentVertex;
						}
						else {
								bHasFinished = true;
								break;
						}
						
						vCalcVals = vCalcVals.yzwx;
				}
				if ( bHasFinished ) { break; }				
		}
		vReturnArray[ iCurrentVertex ] = vReturnArray[ 0 ];
	
		return vReturnArray;
}

// --------------------------------------------------
// Apply 1st Subdivision
// --------------------------------------------------
void Calculate1stSubdivisionNormal( in unsigned int uiVertexID, out vec3 vNormal)
{
		vec2 vTexCoord = vec2( 0.0, 0.0 );
		vec3 vReturnNormal = vec3( 0.0, 0.0, 0.0 );
		
		bool bIsInterior = true;
		int iValence = 0;
		vec3 vOneRing[ciMaxValence] = LoadOneRingOfVertex( uiVertexID, iValence, bIsInterior );		// Loading all vertices surrounding the current (not self)

		vTexCoord = Convert1DIndexTo2DIndex( uiVertexID * unsigned int( iNormalStride ), CalculationSamplerSize );
		vec4 vSaveFirst = texture2D( CalculationSampler, vTexCoord ).xyzw;

		vec3 vVertexSelf = LoadVertex( uiVertexID );												// Loading the current vertex
		vec3 v1 = vec3( 0, 0, 0 );
		vec3 v2 = vec3( 0, 0, 0 );
		
		for ( int i=0; i < iValence - 1; ++i ) {
				v1 = vOneRing[ i   ] - vVertexSelf;
				v2 = vOneRing[ i+1 ] - vVertexSelf;
				vReturnNormal += cross( v1, v2 );
		}		
		
		// interior vertices need to close the gap
		if ( bIsInterior ) {
				v1 = vOneRing[ iValence - 1 ] - vVertexSelf;										// The Last one
				v2 = LoadVertex( unsigned int( vSaveFirst.z ) ) - vVertexSelf;						// where vSaveFirst.z will actually result in vOneRing[0]. But that does not work with GLSL. NVIDIA-techi reading this?
				vReturnNormal += cross( v1, v2 );
		}

		vNormal = normalize( vReturnNormal );
}

// --------------------------------------------------
// Vertex-Shader
// --------------------------------------------------
void main() {
		// the point we instance is 0,0,0 so the + does not actually add something
		// but this way it works without warnings
		unsigned int uiVertexID = unsigned int ( gl_InstanceID ) + unsigned int( gl_Vertex.x );
		vec3 vNormal = vec3(0.0, 0.0, 0.0);
		Calculate1stSubdivisionNormal( uiVertexID, vNormal );
		
		vVarColor = ( vNormal );														// color1 is the normal

		vec2 vPos2D = ConvertOutputPosition( uiVertexID, iOutputSamplerSize );			// The position to write to
		vec4 vPos = vec4(vPos2D.xy, 0.0, 1.0);											// z=0.0, w=1.0
		
		gl_Position = (gl_ModelViewProjectionMatrix*vPos);
}
